# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding field 'Device.health_status'
        db.add_column('lava_scheduler_app_device', 'health_status', self.gf('django.db.models.fields.IntegerField')(default=0), keep_default=False)

        # Adding field 'Device.last_health_report_job'
        db.add_column('lava_scheduler_app_device', 'last_health_report_job', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name=u'Health Report Job', unique=True, null=True, to=orm['lava_scheduler_app.TestJob']), keep_default=False)


    def backwards(self, orm):
        
        # Deleting field 'Device.health_status'
        db.delete_column('lava_scheduler_app_device', 'health_status')

        # Deleting field 'Device.last_health_report_job'
        db.delete_column('lava_scheduler_app_device', 'last_health_report_job_id')


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'lava_scheduler_app.device': {
            'Meta': {'object_name': 'Device'},
            'current_job': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['lava_scheduler_app.TestJob']", 'unique': 'True', 'null': 'True', 'blank': 'True'}),
            'device_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['lava_scheduler_app.DeviceType']"}),
            'health_status': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'hostname': ('django.db.models.fields.CharField', [], {'max_length': '200', 'primary_key': 'True'}),
            'last_health_report_job': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "u'Health Report Job'", 'unique': 'True', 'null': 'True', 'to': "orm['lava_scheduler_app.TestJob']"}),
            'status': ('django.db.models.fields.IntegerField', [], {'default': '1'}),
            'tags': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['lava_scheduler_app.Tag']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'lava_scheduler_app.devicestatetransition': {
            'Meta': {'object_name': 'DeviceStateTransition'},
            'created_by': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'}),
            'created_on': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'device': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'transitions'", 'to': "orm['lava_scheduler_app.Device']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'job': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['lava_scheduler_app.TestJob']", 'null': 'True', 'blank': 'True'}),
            'message': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'new_state': ('django.db.models.fields.IntegerField', [], {}),
            'old_state': ('django.db.models.fields.IntegerField', [], {})
        },
        'lava_scheduler_app.devicetype': {
            'Meta': {'object_name': 'DeviceType'},
            'name': ('django.db.models.fields.SlugField', [], {'max_length': '50', 'primary_key': 'True', 'db_index': 'True'})
        },
        'lava_scheduler_app.tag': {
            'Meta': {'object_name': 'Tag'},
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.SlugField', [], {'unique': 'True', 'max_length': '50', 'db_index': 'True'})
        },
        'lava_scheduler_app.testjob': {
            'Meta': {'object_name': 'TestJob'},
            'actual_device': ('django.db.models.fields.related.ForeignKey', [], {'default': 'None', 'related_name': "'+'", 'null': 'True', 'blank': 'True', 'to': "orm['lava_scheduler_app.Device']"}),
            'definition': ('django.db.models.fields.TextField', [], {}),
            'description': ('django.db.models.fields.CharField', [], {'default': 'None', 'max_length': '200', 'null': 'True', 'blank': 'True'}),
            'end_time': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'log_file': ('django.db.models.fields.files.FileField', [], {'default': 'None', 'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'requested_device': ('django.db.models.fields.related.ForeignKey', [], {'default': 'None', 'related_name': "'+'", 'null': 'True', 'blank': 'True', 'to': "orm['lava_scheduler_app.Device']"}),
            'requested_device_type': ('django.db.models.fields.related.ForeignKey', [], {'default': 'None', 'related_name': "'+'", 'null': 'True', 'blank': 'True', 'to': "orm['lava_scheduler_app.DeviceType']"}),
            'results_link': ('django.db.models.fields.CharField', [], {'default': 'None', 'max_length': '400', 'null': 'True', 'blank': 'True'}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'status': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            'submit_time': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'submit_token': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['linaro_django_xmlrpc.AuthToken']", 'null': 'True', 'blank': 'True'}),
            'submitter': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']"}),
            'tags': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['lava_scheduler_app.Tag']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'linaro_django_xmlrpc.authtoken': {
            'Meta': {'object_name': 'AuthToken'},
            'created_on': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'default': "''", 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'last_used_on': ('django.db.models.fields.DateTimeField', [], {'null': 'True'}),
            'secret': ('django.db.models.fields.CharField', [], {'default': "'wxahi36scmitudt29p0u6qx6ktiq2d0od7uuq0ptklddiiayy1g1yf24p51qlbgbsns74dfdq0bwzz1uh0o4ps5981c2eu0hdrix0smnz4sftf9uros2z002t7gttc5h'", 'unique': 'True', 'max_length': '128'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'auth_tokens'", 'to': "orm['auth.User']"})
        }
    }

    complete_apps = ['lava_scheduler_app']



import logging.config
from optparse import make_option

from django.core.management.base import BaseCommand


class SchedulerCommand(BaseCommand):

    option_list = BaseCommand.option_list + (
        make_option('-l', '--loglevel',
                    action='store',
                    default=None,
                    help="Log level, default is taken from settings."),
        make_option('-f', '--logfile',
                    action='store',
                    default=None,
                    help="Path to log file, default is taken from settings."),
        )

    log_prefix = ''


    _DEFAULT_LOGGING = {
        'version': 1,
        'disable_existing_loggers': True,
        'root': {
            'level': None,
            'handlers': ['default'],
        },
        'formatters': {
            'default': {
                'format': '%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level': 'DEBUG',
                'class': 'logging.FileHandler',
                'formatter': 'verbose'
            }
        },
    }


    def _configure(self, options):
        from django.conf import settings

        daemon_options = settings.SCHEDULER_DAEMON_OPTIONS.copy()
        if options['logfile'] is not None:
            daemon_options['LOG_FILE_PATH'] = options['logfile']
        if options['loglevel'] is not None:
            daemon_options['LOG_LEVEL'] = options['loglevel']

        if daemon_options['LOG_FILE_PATH'] in [None, '-']:
            handler = {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'default',
                }
        else:
            handler = {
                'level': 'DEBUG',
                'class': 'logging.FileHandler',
                'filename': daemon_options['LOG_FILE_PATH'],
                'formatter': 'default'
                }

        fmt = "%(asctime)s [%(levelname)s] [%(name)s] %(message)s"
        if self.log_prefix:
            fmt = self.log_prefix + ' ' + fmt


        LOGGING = {
            'version': 1,
            'disable_existing_loggers': True,
            'root': {
                'level': daemon_options['LOG_LEVEL'].upper(),
                'handlers': ['default'],
            },
            'formatters': {'default': {'format': fmt}},
            'handlers': {'default': handler}
            }

        try:
            import lava.raven
        except ImportError:
            pass
        else:
            LOGGING['handlers']['sentry'] = {
                'level': 'ERROR',
                'class': 'raven.contrib.django.handlers.SentryHandler',
            }
            LOGGING['root']['handlers'].append('sentry')


        logging.config.dictConfig(LOGGING)

        return daemon_options




# Copyright (C) 2011 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of LAVA Scheduler.
#
# LAVA Scheduler is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3 as
# published by the Free Software Foundation
#
# LAVA Scheduler is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA Scheduler.  If not, see <http://www.gnu.org/licenses/>.

import simplejson


from lava_scheduler_app.management.commands import SchedulerCommand
from lava_scheduler_daemon.dbjobsource import DatabaseJobSource


class Command(SchedulerCommand):

    help = "Run the LAVA test job scheduler"

    log_prefix = 'M'

    def handle(self, *args, **options):
        from twisted.internet import reactor
        from lava_scheduler_daemon.board import Job
        daemon_options = self._configure(options)
        source = DatabaseJobSource()
        dispatcher, board_name, json_file = args
        job = Job(
            simplejson.load(open(json_file)), dispatcher,
            source, board_name, reactor, daemon_options=daemon_options)
        def run():
            job.run().addCallback(lambda result: reactor.stop())
        reactor.callWhenRunning(run)
        reactor.run()



from collections import defaultdict
import logging
import os
import simplejson
import StringIO
import datetime
from dateutil.relativedelta import relativedelta

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.core.urlresolvers import reverse
from django.db.models import Count
from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseNotAllowed,
    )
from django.shortcuts import (
    get_object_or_404,
    redirect,
    render_to_response,
)
from django.template import RequestContext
from django.template import defaultfilters as filters
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.db import models

from django_tables2 import Attrs, Column

from lava.utils.data_tables.tables import DataTablesTable

from lava_server.views import index as lava_index
from lava_server.bread_crumbs import (
    BreadCrumb,
    BreadCrumbTrail,
)

from lava_scheduler_app.logfile_helper import (
    formatLogFile,
    getDispatcherErrors,
    getDispatcherLogMessages
    )
from lava_scheduler_app.models import (
    Tag,
    Device,
    DeviceType,
    DeviceStateTransition,
    TestJob,
    )


def post_only(func):
    def decorated(request, *args, **kwargs):
        if request.method != 'POST':
            return HttpResponseNotAllowed('Only POST here')
        return func(request, *args, **kwargs)
    return decorated


class DateColumn(Column):

    def __init__(self, **kw):
        self._format = kw.get('date_format', settings.DATETIME_FORMAT)
        super(DateColumn, self).__init__(**kw)

    def render(self, value):
        return filters.date(value, self._format)


def pklink(record):
    return mark_safe(
        '<a href="%s">%s</a>' % (
            record.get_absolute_url(),
            escape(record.pk)))


class IDLinkColumn(Column):

    def __init__(self, verbose_name="ID", **kw):
        kw['verbose_name'] = verbose_name
        super(IDLinkColumn, self).__init__(**kw)

    def render(self, record):
        return pklink(record)


class RestrictedIDLinkColumn(IDLinkColumn):

    def render(self, record, table):
        if record.is_accessible_by(table.context.get('request').user):
            return pklink(record)
        else:
            return record.pk


def all_jobs_with_device_sort():
    return TestJob.objects.select_related(
        "actual_device", "requested_device", "requested_device_type",
        "submitter", "user", "group").extra(
        select={
            'device_sort': 'coalesce(actual_device_id, requested_device_id, requested_device_type_id)'
            }).all()



class JobTable(DataTablesTable):

    def render_device(self, record):
        if record.actual_device:
            return pklink(record.actual_device)
        elif record.requested_device:
            return pklink(record.requested_device)
        else:
            return mark_safe(
                '<i>' + escape(record.requested_device_type.pk) + '</i>')

    def render_description(self, value):
        if value:
            return value
        else:
            return ''

    id = RestrictedIDLinkColumn()
    status = Column()
    device = Column(accessor='device_sort')
    description = Column(attrs=Attrs(width="30%"))
    submitter = Column()
    submit_time = DateColumn()
    end_time = DateColumn()

    datatable_opts = {
        'aaSorting': [[0, 'desc']],
        }
    searchable_columns=['description']


class IndexJobTable(JobTable):
    def get_queryset(self):
        return all_jobs_with_device_sort().filter(
            status__in=[TestJob.SUBMITTED, TestJob.RUNNING])

    class Meta:
        exclude = ('end_time',)


def index_active_jobs_json(request):
    return IndexJobTable.json(request)


class DeviceTable(DataTablesTable):

    def get_queryset(self):
        return Device.objects.select_related("device_type")

    hostname = IDLinkColumn("hostname")
    device_type = Column()
    status = Column()
    health_status = Column()

    searchable_columns=['hostname']


def index_devices_json(request):
    return DeviceTable.json(request)

def health_jobs_in_hr(hr=-24):
    return TestJob.objects.filter(health_check=True,
           start_time__gte=(datetime.datetime.now()
               + relativedelta(hours=hr))).exclude(status__in=[TestJob.SUBMITTED, TestJob.RUNNING])

def _online_total():
    ''' returns a tuple of (num_online, num_not_retired) '''
    r = Device.objects.all().values('status').annotate(count=Count('status'))
    offline = total = 0
    for res in r:
        if res['status'] in [Device.OFFLINE, Device.OFFLINING]:
            offline += res['count']
        if res['status'] != Device.RETIRED:
            total += res['count']

    return (total-offline,total)

@BreadCrumb("Scheduler", parent=lava_index)
def index(request):
    return render_to_response(
        "lava_scheduler_app/index.html",
        {
            'device_status': "%d/%d" % _online_total(),
            'health_check_status': "%s/%s" % (
                health_jobs_in_hr().filter(status=TestJob.COMPLETE).count(),
                health_jobs_in_hr().count()),
            'device_type_table': DeviceTypeTable('devicetype', reverse(device_type_json)),
            'devices_table': DeviceTable('devices', reverse(index_devices_json)),
            'active_jobs_table': IndexJobTable(
                'active_jobs', reverse(index_active_jobs_json)),
            'bread_crumb_trail': BreadCrumbTrail.leading_to(index),
        },
        RequestContext(request))

def job_report(start_day, end_day, health_check):
    now = datetime.datetime.now()
    start_date = now + datetime.timedelta(start_day)
    end_date = now + datetime.timedelta(end_day)

    res = TestJob.objects.filter(
            health_check=health_check,
            start_time__range=(start_date, end_date),
            status__in=(
                TestJob.COMPLETE,
                TestJob.INCOMPLETE, TestJob.CANCELED, TestJob.CANCELING),
        ).values(
            'status'
        )
    return {
        'pass': res.filter(status=TestJob.COMPLETE).count(),
        'fail': res.exclude(status=TestJob.COMPLETE).count(),
        'date': start_date.strftime('%m-%d'),
    }

@BreadCrumb("Reports", parent=lava_index)
def reports(request):
    health_day_report = []
    health_week_report = []
    job_day_report = []
    job_week_report = []
    for day in reversed(range(7)):
        health_day_report.append(job_report(day*-1-1, day*-1, True))
        job_day_report.append(job_report(day*-1-1, day*-1, False))
    for week in reversed(range(10)):
        health_week_report.append(job_report(week*-7-7, week*-7, True))
        job_week_report.append(job_report(week*-7-7, week*-7, False))
    return render_to_response(
        "lava_scheduler_app/reports.html",
        {
            'health_week_report': health_week_report,
            'health_day_report': health_day_report,
            'job_week_report': job_week_report,
            'job_day_report': job_day_report,
            'bread_crumb_trail': BreadCrumbTrail.leading_to(index),
        },
        RequestContext(request))

@BreadCrumb("All Devices", parent=index)
def device_list(request):
    return render_to_response(
        "lava_scheduler_app/alldevices.html",
        {
            'devices_table': DeviceTable('devices', reverse(index_devices_json)),
            'bread_crumb_trail': BreadCrumbTrail.leading_to(device_list),
        },
        RequestContext(request))

def get_restricted_job(user, pk):
    job =  get_object_or_404(TestJob.objects, pk=pk)
    if not job.is_accessible_by(user):
        raise PermissionDenied()
    return job

class SumIfSQL(models.sql.aggregates.Aggregate):
    is_ordinal = True
    sql_function = 'SUM'
    sql_template = 'SUM((%(condition)s)::int)'

class SumIf(models.Aggregate):
    name = 'SumIf'
    def add_to_query(self, query, alias, col, source, is_summary):
        aggregate = SumIfSQL(
                col, source=source, is_summary=is_summary, **self.extra)
        query.aggregates[alias] = aggregate

class DeviceTypeTable(DataTablesTable):

    def get_queryset(self):
        return DeviceType.objects.all().annotate(
            idle=SumIf('device', condition='status=%s' % Device.IDLE),
            offline=SumIf('device', condition='status in (%s,%s)' % (
                Device.OFFLINE, Device.OFFLINING)),
            busy=SumIf('device', condition='status=%s' % Device.RUNNING),
            ).order_by('name')

    def render_status(self, record):
        return "%s idle, %s offline, %s busy" % (record.idle, record.offline,
                record.busy)

    name = IDLinkColumn("name")
    status = Column()

    searchable_columns = ['name']


class HealthJobSummaryTable(DataTablesTable):
    """
    The Table will return 1 day, 1 week, 1 month offset health job count.
    The value is defined when table instance is created in device_type_detail()
    """

    def render_name(self, record):
        matrix = {-24:"24hours", -24*7:"Week", -24*7*30:"Month"}
        return matrix[record]

    def render_Complete(self, record):
        device_type = self.params[0]
        num = health_jobs_in_hr(record).filter(
                actual_device__in=Device.objects.filter(
                device_type=device_type), status=TestJob.COMPLETE).count()
        return num

    def render_Failed(self, record):
        device_type = self.params[0]
        num = health_jobs_in_hr(record).filter(
                actual_device__in=Device.objects.filter(
                device_type=device_type), status__in=[TestJob.INCOMPLETE,
                    TestJob.CANCELED, TestJob.CANCELING]).count()
        return num

    name = Column()
    Complete = Column()
    Failed = Column()

def device_type_json(request):
    return DeviceTypeTable.json(request)

class NoDTDeviceTable(DeviceTable):
    def get_queryset(self, device_type):
        return Device.objects.filter(device_type=device_type)

    class Meta:
        exclude = ('device_type',)

def index_nodt_devices_json(request, pk):
    device_type = get_object_or_404(DeviceType, pk=pk)
    return NoDTDeviceTable.json(request, params=(device_type,))

@BreadCrumb("Device Type {pk}", parent=index, needs=['pk'])
def device_type_detail(request, pk):
    dt = get_object_or_404(DeviceType, pk=pk)
    return render_to_response(
        "lava_scheduler_app/device_type.html",
        {
            'device_type': dt,
            'running_jobs_num': TestJob.objects.filter(
                actual_device__in=Device.objects.filter(device_type=dt),
                status=TestJob.RUNNING).count(),
            # Fix me: doesn't count actual_device not set but requested
            # device type jobs.
            'queued_jobs_num': TestJob.objects.filter(
                actual_device__in=Device.objects.filter(device_type=dt),
                status=TestJob.SUBMITTED).count(),
            # data return 1 day, 1 week, 1 month offset
            'health_job_summary_table': HealthJobSummaryTable(
                'device_type', params=(dt,), data=[-24, -24*7, -24*7*30]),
            'devices_table_no_dt': NoDTDeviceTable('devices',
                reverse(index_nodt_devices_json, kwargs=dict(pk=pk)), params=(dt,)),
            'bread_crumb_trail': BreadCrumbTrail.leading_to(device_type_detail, pk=pk),
        },
        RequestContext(request))


class DeviceHealthTable(DataTablesTable):

    def get_queryset(self):
        return Device.objects.select_related(
            "hostname", "last_health_report_job")

    def render_hostname(self, record):
        return mark_safe('<a href="%s">%s</a>' % (
            record.get_device_health_url(), escape(record.pk)))

    def render_last_health_report_job(self, record):
        report = record.last_health_report_job
        if report is None:
            return ''
        else:
            return pklink(report)

    hostname = Column("hostname")
    health_status = Column()
    last_report_time = DateColumn(
        verbose_name="last report time",
        accessor="last_health_report_job.end_time")
    last_health_report_job = Column("last report job")

    searchable_columns=['hostname']
    datatable_opts = {
        "iDisplayLength": 25
        }


def lab_health_json(request):
    return DeviceHealthTable.json(request)


@BreadCrumb("All Device Health", parent=index)
def lab_health(request):
    return render_to_response(
        "lava_scheduler_app/labhealth.html",
        {
            'device_health_table': DeviceHealthTable(
                'device_health', reverse(lab_health_json)),
            'bread_crumb_trail': BreadCrumbTrail.leading_to(lab_health),
        },
        RequestContext(request))


class HealthJobTable(JobTable):

    def get_queryset(self, device):
        return TestJob.objects.select_related(
            "submitter",
            ).filter(
            actual_device=device,
            health_check=True)

    class Meta:
        exclude = ('description', 'device')


def health_jobs_json(request, pk):
    device = get_object_or_404(Device, pk=pk)
    return HealthJobTable.json(params=(device,))


@BreadCrumb("All Health Jobs on Device {pk}", parent=index, needs=['pk'])
def health_job_list(request, pk):
    device = get_object_or_404(Device, pk=pk)

    return render_to_response(
        "lava_scheduler_app/health_jobs.html",
        {
            'device': device,
            'transition_table': DeviceTransitionTable(
                'transitions', reverse(transition_json, kwargs=dict(pk=device.pk)),
                params=(device,)),
            'health_job_table': HealthJobTable(
                'health_jobs', reverse(health_jobs_json, kwargs=dict(pk=pk)),
                params=(device,)),
            'show_maintenance': device.can_admin(request.user) and \
                device.status in [Device.IDLE, Device.RUNNING],
            'show_online': device.can_admin(request.user) and \
                device.status in [Device.OFFLINE, Device.OFFLINING],
            'bread_crumb_trail': BreadCrumbTrail.leading_to(health_job_list, pk=pk),
        },
        RequestContext(request))


class AllJobsTable(JobTable):

    def get_queryset(self):
        return all_jobs_with_device_sort()

    datatable_opts = JobTable.datatable_opts.copy()

    datatable_opts.update({
        'iDisplayLength': 25,
        })


def alljobs_json(request):
    return AllJobsTable.json(request)


@BreadCrumb("All Jobs", parent=index)
def job_list(request):
    return render_to_response(
        "lava_scheduler_app/alljobs.html",
        {
            'bread_crumb_trail': BreadCrumbTrail.leading_to(job_list),
            'alljobs_table': AllJobsTable('alljobs', reverse(alljobs_json)),
        },
        RequestContext(request))


@BreadCrumb("Job #{pk}", parent=index, needs=['pk'])
def job_detail(request, pk):
    job = get_restricted_job(request.user, pk)

    data = {
        'job': job,
        'show_cancel': job.status <= TestJob.RUNNING and job.can_cancel(request.user),
        'bread_crumb_trail': BreadCrumbTrail.leading_to(job_detail, pk=pk),
        'show_reload_page' : job.status <= TestJob.RUNNING,
    }

    if job.log_file:
        job_errors = getDispatcherErrors(job.log_file)
        job_log_messages = getDispatcherLogMessages(job.log_file)

        levels = defaultdict(int)
        for kl in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']:
            levels[kl] = 0
        for level, msg, _ in job_log_messages:
            levels[level] += 1
        levels = sorted(levels.items(), key=lambda (k,v):logging._levelNames.get(k))
        data.update({
            'job_file_present': True,
            'job_errors' : job_errors,
            'job_has_error' : len(job_errors) > 0,
            'job_log_messages' : job_log_messages,
            'levels': levels,
            'job_file_size' : job.log_file.size,
            })
    else:
        data.update({
            'job_file_present': False,
            })

    return render_to_response(
        "lava_scheduler_app/job.html", data, RequestContext(request))


def job_definition(request, pk):
    job = get_restricted_job(request.user, pk)
    return render_to_response(
        "lava_scheduler_app/job_definition.html",
        {
            'job': job,
            'job_file_present': bool(job.log_file),
        },
        RequestContext(request))


def job_definition_plain(request, pk):
    job = get_restricted_job(request.user, pk)
    response = HttpResponse(job.definition, mimetype='text/plain')
    response['Content-Disposition'] = "attachment; filename=job_%d.json"%job.id
    return response


@BreadCrumb("Complete log", parent=job_detail, needs=['pk'])
def job_log_file(request, pk):
    job = get_restricted_job(request.user, pk)
    content = formatLogFile(job.log_file)
    return render_to_response(
        "lava_scheduler_app/job_log_file.html",
        {
            'job': TestJob.objects.get(pk=pk),
            'job_file_present': bool(job.log_file),
            'sections' : content,
            'job_file_size' : job.log_file.size,
        },
        RequestContext(request))


def job_log_file_plain(request, pk):
    job = get_restricted_job(request.user, pk)
    response = HttpResponse(job.log_file, mimetype='text/plain')
    response['Content-Disposition'] = "attachment; filename=job_%d.log"%job.id
    return response


def job_log_incremental(request, pk):
    start = int(request.GET.get('start', 0))
    job = get_restricted_job(request.user, pk)
    log_file = job.log_file
    log_file.seek(start)
    new_content = log_file.read()
    m = getDispatcherLogMessages(StringIO.StringIO(new_content))
    response = HttpResponse(
        simplejson.dumps(m), content_type='application/json')
    response['X-Current-Size'] = str(start + len(new_content))
    if job.status not in [TestJob.RUNNING, TestJob.CANCELING]:
        response['X-Is-Finished'] = '1'
    return response


def job_full_log_incremental(request, pk):
    start = int(request.GET.get('start', 0))
    job = get_restricted_job(request.user, pk)
    log_file = job.log_file
    log_file.seek(start)
    new_content = log_file.read()
    nl_index = new_content.rfind('\n', -NEWLINE_SCAN_SIZE)
    if nl_index >= 0:
        new_content = new_content[:nl_index+1]
    m = formatLogFile(StringIO.StringIO(new_content))
    response = HttpResponse(
        simplejson.dumps(m), content_type='application/json')
    response['X-Current-Size'] = str(start + len(new_content))
    if job.status not in [TestJob.RUNNING, TestJob.CANCELING]:
        response['X-Is-Finished'] = '1'
    return response


LOG_CHUNK_SIZE = 512*1024
NEWLINE_SCAN_SIZE = 80


def job_output(request, pk):
    start = request.GET.get('start', 0)
    try:
        start = int(start)
    except ValueError:
        return HttpResponseBadRequest("invalid start")
    count_present = 'count' in request.GET
    job = get_restricted_job(request.user, pk)
    log_file = job.log_file
    log_file.seek(0, os.SEEK_END)
    size = int(request.GET.get('count', log_file.tell()))
    if size - start > LOG_CHUNK_SIZE and not count_present:
        log_file.seek(-LOG_CHUNK_SIZE, os.SEEK_END)
        content = log_file.read(LOG_CHUNK_SIZE)
        nl_index = content.find('\n', 0, NEWLINE_SCAN_SIZE)
        if nl_index > 0 and not count_present:
            content = content[nl_index + 1:]
        skipped = size - start - len(content)
    else:
        skipped = 0
        log_file.seek(start, os.SEEK_SET)
        content = log_file.read(size - start)
    nl_index = content.rfind('\n', -NEWLINE_SCAN_SIZE)
    if nl_index >= 0 and not count_present:
        content = content[:nl_index+1]
    response = HttpResponse(content)
    if skipped:
        response['X-Skipped-Bytes'] = str(skipped)
    response['X-Current-Size'] = str(start + len(content))
    if job.status not in [TestJob.RUNNING, TestJob.CANCELING]:
        response['X-Is-Finished'] = '1'
    return response


@post_only
def job_cancel(request, pk):
    job = get_restricted_job(request.user, pk)
    if job.can_cancel(request.user):
        job.cancel()
        return redirect(job)
    else:
        return HttpResponseForbidden(
            "you cannot cancel this job", content_type="text/plain")


def job_json(request, pk):
    job = get_restricted_job(request.user, pk)
    json_text = simplejson.dumps({
        'status': job.get_status_display(),
        'results_link': job.results_link,
        })
    content_type = 'application/json'
    if 'callback' in request.GET:
        json_text = '%s(%s)'%(request.GET['callback'], json_text)
        content_type = 'text/javascript'
    return HttpResponse(json_text, content_type=content_type)


class RecentJobsTable(JobTable):

    def get_queryset(self, device):
        return device.recent_jobs()

    class Meta:
        exclude = ('device',)


def recent_jobs_json(request, pk):
    device = get_object_or_404(Device, pk=pk)
    return RecentJobsTable.json(request, params=(device,))


class DeviceTransitionTable(DataTablesTable):

    def get_queryset(self, device):
        qs = device.transitions.select_related('created_by')
        qs = qs.extra(select={'prev': """
        select t.created_on
          from lava_scheduler_app_devicestatetransition as t
         where t.device_id=%s and t.created_on < lava_scheduler_app_devicestatetransition.created_on
         order by t.created_on desc
         limit 1 """},
                      select_params=[device.pk])
        return qs

    def render_created_on(self, record):
        t = record
        base = filters.date(t.created_on, "Y-m-d H:i")
        if t.prev:
            base += ' (after %s)' % (filters.timesince(t.prev, t.created_on))
        return base

    def render_transition(self, record):
        t = record
        return mark_safe(
            '%s &rarr; %s' % (t.get_old_state_display(), t.get_new_state_display(),))

    def render_message(self, value):
        if value is None:
            return ''
        else:
            return value

    created_on = Column('when', attrs=Attrs(width="40%"))
    transition = Column('transition', sortable=False)
    created_by = Column('by')
    message = Column('reason')

    datatable_opts = {
        'aaSorting': [[0, 'desc']],
        }


def transition_json(request, pk):
    device = get_object_or_404(Device, pk=pk)
    return DeviceTransitionTable.json(request, params=(device,))


@BreadCrumb("Device {pk}", parent=index, needs=['pk'])
def device_detail(request, pk):
    device = get_object_or_404(Device, pk=pk)
    if device.status in [Device.OFFLINE, Device.OFFLINING]:
        try:
            transition = device.transitions.filter(message__isnull=False).latest('created_on').message
        except DeviceStateTransition.DoesNotExist:
            transition = None
    else:
        transition = None
    return render_to_response(
        "lava_scheduler_app/device.html",
        {
            'device': device,
            'transition': transition,
            'transition_table': DeviceTransitionTable(
                'transitions', reverse(transition_json, kwargs=dict(pk=device.pk)),
                params=(device,)),
            'recent_job_table': RecentJobsTable(
                'jobs', reverse(recent_jobs_json, kwargs=dict(pk=device.pk)),
                params=(device,)),
            'show_maintenance': device.can_admin(request.user) and \
                device.status in [Device.IDLE, Device.RUNNING],
            'show_online': device.can_admin(request.user) and \
                device.status in [Device.OFFLINE, Device.OFFLINING],
            'bread_crumb_trail': BreadCrumbTrail.leading_to(device_detail, pk=pk),
        },
        RequestContext(request))


@post_only
def device_maintenance_mode(request, pk):
    device = Device.objects.get(pk=pk)
    if device.can_admin(request.user):
        device.put_into_maintenance_mode(request.user, request.POST.get('reason'))
        return redirect(device)
    else:
        return HttpResponseForbidden(
            "you cannot administer this device", content_type="text/plain")


@post_only
def device_online(request, pk):
    device = Device.objects.get(pk=pk)
    if device.can_admin(request.user):
        device.put_into_online_mode(request.user, request.POST.get('reason'))
        return redirect(device)
    else:
        return HttpResponseForbidden(
            "you cannot administer this device", content_type="text/plain")

@post_only
def device_looping_mode(request, pk):
    device = Device.objects.get(pk=pk)
    if device.can_admin(request.user):
        device.put_into_looping_mode(request.user)
        return redirect(device)
    else:
        return HttpResponseForbidden(
            "you cannot administer this device", content_type="text/plain")




from collections import defaultdict
import logging

from twisted.internet import defer
from twisted.internet.task import Clock
from twisted.trial.unittest import TestCase

from lava_scheduler_daemon.board import Board

def stub_method(method_name):
    def method_impl(self, board_name, *args):
        assert method_name not in self._requests[board_name], (
            'overlapping call to %s on %s' % (method_name, board_name))
        d = self._requests[method_name][board_name] = defer.Deferred()
        def _remove_request(result):
            del self._requests[method_name][board_name]
            return result
        d.addBoth(_remove_request)
        self._calls[board_name][method_name].append(args)
        return d
    return method_impl


class TestJobSource(object):

    def __init__(self):
        self._calls = defaultdict(lambda :defaultdict(list))
        self._requests = defaultdict(dict)

    jobCompleted = stub_method('jobCompleted')
    getJobForBoard = stub_method('getJobForBoard')

    def _completeCall(self, method_name, board_name, result):
        self._requests[method_name][board_name].callback(result)


class TestJob(object):

    def __init__(self, job_data, dispatcher, source, board_name, reactor, options):
        self.json_data = job_data
        self.dispatcher = dispatcher
        self.reactor = reactor
        self.source = source
        self.board_name = board_name
        self.deferred = defer.Deferred()

    def run(self):
        return self.deferred


class AppendingHandler(logging.Handler):

    def __init__(self, target_list):
        logging.Handler.__init__(self)
        self.target_list = target_list

    def emit(self, record):
        self.target_list.append((record.levelno, self.format(record)))


class TestBoard(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        self.clock = Clock()
        self.source = TestJobSource()
        self._log_messages = []
        self._handler = AppendingHandler(self._log_messages)
        self.addCleanup(self._checkNoLogs)

    def _checkNoLogs(self):
        warnings = [message for (level, message) in self._log_messages
                    if level >= logging.WARNING]
        if warnings:
            self.fail("Logged warnings: %s" % warnings)

    def make_board(self, board_name):
        board = Board(
            self.source, board_name, 'script', self.clock, None, job_cls=TestJob)
        board.logger.addHandler(self._handler)
        board.logger.setLevel(logging.DEBUG)
        return board

    def test_initial_state_is_stopped(self):
        b = self.make_board('board')
        self.assertEqual('S', b._state_name())

    def test_start_checks(self):
        b = self.make_board('board')
        b.start()
        self.assertEqual('C', b._state_name())

    def test_no_job_waits(self):
        b = self.make_board('board')
        b.start()
        self.source._completeCall('getJobForBoard', 'board', None)
        self.assertEqual('W', b._state_name())

    def test_actual_job_runs(self):
        b = self.make_board('board')
        b.start()
        self.source._completeCall('getJobForBoard', 'board', ({}, None))
        self.assertEqual('R', b._state_name())

    def test_check_again_on_completion(self):
        b = self.make_board('board')
        b.start()
        self.source._completeCall('getJobForBoard', 'board', ({}, None))
        b.running_job.deferred.callback('path')
        self.assertEqual('C', b._state_name())

    def test_stop_while_checking_moves_to_check_plus_stop(self):
        b = self.make_board('board')
        b.start()
        b.stop()
        self.assertEqual('C+S', b._state_name())

    def test_stop_while_checking_no_job_stops(self):
        b = self.make_board('board')
        b.start()
        s = b.stop()
        stop_results = []
        s.addCallback(stop_results.append)
        self.assertEqual(0, len(stop_results))
        self.source._completeCall('getJobForBoard', 'board', None)
        self.assertEqual(1, len(stop_results))
        self.assertEqual('S', b._state_name())

    def test_stop_while_checking_actual_job_runs(self):
        b = self.make_board('board')
        b.start()
        s = b.stop()
        stop_results = []
        s.addCallback(stop_results.append)
        self.assertEqual(0, len(stop_results))
        self.source._completeCall('getJobForBoard', 'board', ({}, None))
        self.assertEqual(0, len(stop_results))
        self.assertEqual('R+S', b._state_name())

    def test_stop_while_checking_actual_job_stops_on_complete(self):
        b = self.make_board('board')
        b.start()
        s = b.stop()
        stop_results = []
        s.addCallback(stop_results.append)
        self.assertEqual(0, len(stop_results))
        self.source._completeCall('getJobForBoard', 'board', ({}, None))
        b.running_job.deferred.callback(None)
        self.assertEqual(1, len(stop_results))
        self.assertEqual('S', b._state_name())

    def test_stop_while_running_job_stops_on_complete(self):
        b = self.make_board('board')
        b.start()
        self.source._completeCall('getJobForBoard', 'board', ({}, None))
        self.assertEqual('R', b._state_name())
        s = b.stop()
        stop_results = []
        s.addCallback(stop_results.append)
        self.assertEqual(0, len(stop_results))
        b.running_job.deferred.callback(None)
        self.assertEqual(1, len(stop_results))
        self.assertEqual('S', b._state_name())

    def test_wait_expires_check_again(self):
        b = self.make_board('board')
        b.start()
        self.source._completeCall('getJobForBoard', 'board', None)
        self.clock.advance(10000) # hack: the delay should be config data
        self.assertEqual('C', b._state_name())



import json
import logging

from twisted.internet import defer

from zope.interface import (
    implements,
    Interface,
    )

logger = logging.getLogger(__name__)


class IJobSource(Interface):

    def getBoardList():
        """Get the list of currently configured board names."""

    def getJobForBoard(board_name):
        """Return the json data of a job for board_name and a log file.

        The job should be marked as started before it is returned.
        """

    def jobCompleted(board_name):
        """Mark the job currently running on `board_name` as completed."""


class DirectoryJobSource(object):

    implements(IJobSource)

    logger = logging.getLogger(__name__ + '.DirectoryJobSource')

    def __init__(self, directory):
        self.directory = directory
        if not self.directory.isdir():
            self.logger.critical("%s is not a directory", self.directory)
            raise RuntimeError("%s must be a directory" % self.directory)
        boards = self.directory.child('boards')
        if not boards.isdir():
            self.logger.critical("%s is not a directory", boards)
            raise RuntimeError("%s must be a directory" % boards)
        for subdir in 'incoming', 'completed', 'broken':
            subdir = self.directory.child(subdir)
            if not subdir.isdir():
                subdir.createDirectory()
        self.logger.info("starting to look for jobs in %s", self.directory)

    def _getBoardList(self):
        return self.directory.child('boards').listdir()

    def getBoardList(self):
        return defer.maybeDeferred(self._getBoardList)

    def _jsons(self, kind):
        files = self.directory.child(kind).globChildren("*.json")
        for json_file in files:
            yield (json.load(json_file.open()), json_file)

    def _board_dir(self, board_name):
        return self.directory.child('boards').child(board_name)

    def _getJobForBoard(self, board_name):
        self.logger.debug('getting job for %s', board_name)
        board_dir = self._board_dir(board_name)
        if board_dir.listdir() != []:
            self.logger.debug('board %s busy', board_name)
            return None
        for json_data, json_file in self._jsons('incoming'):
            self.logger.debug('considering %s for %s', json_file, board_name)
            if json_data['target'] == board_name:
                self.logger.debug('running %s on %s', json_file, board_name)
                json_file.moveTo(board_dir.child(json_file.basename()))
                return json_data, open('/dev/null', 'w')
        else:
            return None

    def getJobForBoard(self, board_name):
        return defer.maybeDeferred(self._getJobForBoard, board_name)

    def _jobCompleted(self, board_name):
        [json_file] = self._board_dir(board_name).children()
        completed = self.directory.child('completed')
        counter = 0
        while True:
            fname = '%03d%s' % (counter, json_file.basename())
            if not completed.child(fname).exists():
                break
            counter += 1
        json_file.moveTo(completed.child(fname))

    def jobCompleted(self, board_name):
        return defer.maybeDeferred(self._jobCompleted, board_name)



import cStringIO
import datetime
import json
import xmlrpclib

from dashboard_app.models import BundleStream

from django.contrib.auth.models import Group, Permission, User
from django.test import TransactionTestCase
from django.test.client import Client

from django_testscenarios.ubertest import TestCase

from linaro_django_xmlrpc.models import AuthToken

from lava_scheduler_app.models import (
    Device,
    DeviceType,
    JSONDataError,
    Tag,
    TestJob)
from lava_scheduler_daemon.dbjobsource import DatabaseJobSource


# Based on http://www.technobabble.dk/2008/apr/02/xml-rpc-dispatching-through-django-test-client/
class TestTransport(xmlrpclib.Transport):
    """Handles connections to XML-RPC server through Django test client."""

    def __init__(self, user=None, password=None):
        self.client = Client()
        if user:
            success = self.client.login(username=user, password=password)
            if not success:
                raise AssertionError("Login attempt failed!")
        self._use_datetime = True

    def request(self, host, handler, request_body, verbose=0):
        self.verbose = verbose
        response = self.client.post(
            handler, request_body, content_type="text/xml")
        res = cStringIO.StringIO(response.content)
        res.seek(0)
        return self.parse_response(res)


class ModelFactory(object):

    def __init__(self):
        self._int = 0

    def getUniqueInteger(self):
        self._int += 1
        return self._int

    def getUniqueString(self, prefix='generic'):
        return '%s-%d' % (prefix, self.getUniqueInteger())

    def make_user(self):
        return User.objects.create_user(
            self.getUniqueString(),
            '%s@mail.invalid' % (self.getUniqueString(),),
            self.getUniqueString())

    def ensure_device_type(self, name=None):
        if name is None:
            name = self.getUniqueString('name')
        return DeviceType.objects.get_or_create(name=name)[0]

    def make_device_type(self, name=None, health_check_job=None):
        if name is None:
            name = self.getUniqueString('name')
        device_type = DeviceType.objects.create(
            name=name, health_check_job=health_check_job)
        device_type.save()
        return device_type

    def ensure_tag(self, name):
        return Tag.objects.get_or_create(name=name)[0]

    def make_device(self, device_type=None, hostname=None, **kw):
        if device_type is None:
            device_type = self.ensure_device_type()
        if hostname is None:
            hostname = self.getUniqueString()
        device = Device(device_type=device_type, hostname=hostname, **kw)
        device.save()
        return device

    def make_job_data(self, actions=[], **kw):
        data = {'actions': actions, 'timeout': 1}
        data.update(kw)
        if 'target' not in data and 'device_type' not in data:
            if DeviceType.objects.all():
                data['device_type'] = DeviceType.objects.all()[0].name
            else:
                data['device_type'] = self.ensure_device_type().name
        return data

    def make_job_json(self, **kw):
        return json.dumps(self.make_job_data(**kw))

    def make_testjob(self, definition=None, submitter=None, **kwargs):
        if definition is None:
            definition = self.make_job_json()
        if submitter is None:
            submitter = self.make_user()
        if 'user' not in kwargs:
            kwargs['user'] = submitter
        testjob = TestJob(
            definition=definition, submitter=submitter, **kwargs)
        testjob.save()
        return testjob


class TestCaseWithFactory(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        self.factory = ModelFactory()


class TestTestJob(TestCaseWithFactory):

    def test_from_json_and_user_sets_definition(self):
        definition = self.factory.make_job_json()
        job = TestJob.from_json_and_user(definition, self.factory.make_user())
        self.assertEqual(definition, job.definition)

    def test_from_json_and_user_sets_submitter(self):
        user = self.factory.make_user()
        job = TestJob.from_json_and_user(
            self.factory.make_job_json(), user)
        self.assertEqual(user, job.submitter)

    def test_from_json_and_user_sets_device_type(self):
        panda_type = self.factory.ensure_device_type(name='panda')
        job = TestJob.from_json_and_user(
            self.factory.make_job_json(device_type='panda'),
            self.factory.make_user())
        self.assertEqual(panda_type, job.requested_device_type)

    def test_from_json_and_user_sets_target(self):
        panda_board = self.factory.make_device(hostname='panda01')
        job = TestJob.from_json_and_user(
            self.factory.make_job_json(target='panda01'),
            self.factory.make_user())
        self.assertEqual(panda_board, job.requested_device)

    def test_from_json_and_user_does_not_set_device_type_from_target(self):
        panda_type = self.factory.ensure_device_type(name='panda')
        self.factory.make_device(device_type=panda_type, hostname='panda01')
        job = TestJob.from_json_and_user(
            self.factory.make_job_json(target='panda01'),
            self.factory.make_user())
        self.assertEqual(None, job.requested_device_type)

    def test_from_json_and_user_sets_date_submitted(self):
        before = datetime.datetime.now()
        job = TestJob.from_json_and_user(
            self.factory.make_job_json(),
            self.factory.make_user())
        after = datetime.datetime.now()
        self.assertTrue(before < job.submit_time < after)

    def test_from_json_and_user_sets_status_to_SUBMITTED(self):
        job = TestJob.from_json_and_user(
            self.factory.make_job_json(),
            self.factory.make_user())
        self.assertEqual(job.status, TestJob.SUBMITTED)

    def test_from_json_and_user_sets_no_tags_if_no_tags(self):
        job = TestJob.from_json_and_user(
            self.factory.make_job_json(device_tags=[]),
            self.factory.make_user())
        self.assertEqual(set(job.tags.all()), set([]))

    def test_from_json_and_user_errors_on_unknown_tags(self):
        self.assertRaises(
            JSONDataError, TestJob.from_json_and_user,
            self.factory.make_job_json(device_tags=['unknown']),
            self.factory.make_user())

    def test_from_json_and_user_sets_tag_from_device_tags(self):
        self.factory.ensure_tag('tag')
        job = TestJob.from_json_and_user(
            self.factory.make_job_json(device_tags=['tag']),
            self.factory.make_user())
        self.assertEqual(
            set(tag.name for tag in job.tags.all()), set(['tag']))

    def test_from_json_and_user_sets_multiple_tag_from_device_tags(self):
        self.factory.ensure_device_type(name='panda')
        self.factory.ensure_tag('tag1')
        self.factory.ensure_tag('tag2')
        job = TestJob.from_json_and_user(
            self.factory.make_job_json(device_tags=['tag1', 'tag2']),
            self.factory.make_user())
        self.assertEqual(
            set(tag.name for tag in job.tags.all()), set(['tag1', 'tag2']))

    def test_from_json_and_user_reuses_tag_objects(self):
        self.factory.ensure_device_type(name='panda')
        self.factory.ensure_tag('tag')
        job1 = TestJob.from_json_and_user(
            self.factory.make_job_json(device_tags=['tag']),
            self.factory.make_user())
        job2 = TestJob.from_json_and_user(
            self.factory.make_job_json(device_tags=['tag']),
            self.factory.make_user())
        self.assertEqual(
            set(tag.pk for tag in job1.tags.all()),
            set(tag.pk for tag in job2.tags.all()))

    def test_from_json_and_user_rejects_invalid_json(self):
        self.assertRaises(
            ValueError, TestJob.from_json_and_user, '{',
            self.factory.make_user())

    def test_from_json_and_user_rejects_invalid_job(self):
        # job data must have the 'actions' and 'timeout' properties, so this
        # will be rejected.
        self.assertRaises(
            ValueError, TestJob.from_json_and_user, '{}',
            self.factory.make_user())

    def make_job_json_for_stream_name(self, stream_name):
        return self.factory.make_job_json(
            actions=[
                {
                    'command':'submit_results',
                    'parameters': {
                        'server': '...',
                        'stream': stream_name,
                        }
                    }
                ])

    def test_from_json_and_user_sets_group_from_bundlestream(self):
        group = Group.objects.create(name='group')
        user = self.factory.make_user()
        user.groups.add(group)
        b = BundleStream.objects.create(
            group=group, slug='blah', is_public=True)
        b.save()
        j = self.make_job_json_for_stream_name(b.pathname)
        job = TestJob.from_json_and_user(j, user)
        self.assertEqual(group, job.group)

    def test_from_json_and_user_can_submit_to_anonymous(self):
        user = self.factory.make_user()
        anon_user = User.objects.get_or_create(username="anonymous-owner")[0]
        b = BundleStream.objects.create(
            slug='anonymous', is_anonymous=True, user=anon_user,
            is_public=True)
        b.save()
        j = self.make_job_json_for_stream_name('/anonymous/anonymous/')
        job = TestJob.from_json_and_user(j, user)
        self.assertEqual(user, job.submitter)

    def test_from_json_and_user_sets_is_public_from_bundlestream(self):
        group = Group.objects.create(name='group')
        user = self.factory.make_user()
        user.groups.add(group)
        b = BundleStream.objects.create(
            group=group, slug='blah', is_public=False)
        b.save()
        j = self.make_job_json_for_stream_name(b.pathname)
        job = TestJob.from_json_and_user(j, user)
        self.assertEqual(False, job.is_public)

    def test_from_json_and_user_rejects_missing_bundlestream(self):
        user = self.factory.make_user()
        j = self.make_job_json_for_stream_name('no such stream')
        self.assertRaises(ValueError, TestJob.from_json_and_user, j, user)

    def test_from_json_and_user_rejects_inaccessible_bundlestream(self):
        stream_user = self.factory.make_user()
        job_user = self.factory.make_user()
        b = BundleStream.objects.create(
            user=stream_user, slug='blah', is_public=True)
        b.save()
        j = self.make_job_json_for_stream_name(b.pathname)
        self.assertRaises(ValueError, TestJob.from_json_and_user, j, job_user)


class TestSchedulerAPI(TestCaseWithFactory):

    def server_proxy(self, user=None, password=None):
        return xmlrpclib.ServerProxy(
            'http://localhost/RPC2/',
            transport=TestTransport(user=user, password=password))

    def test_submit_job_rejects_anonymous(self):
        server = self.server_proxy()
        try:
            server.scheduler.submit_job("{}")
        except xmlrpclib.Fault as f:
            self.assertEqual(401, f.faultCode)
        else:
            self.fail("fault not raised")

    def test_submit_job_rejects_unpriv_user(self):
        User.objects.create_user('test', 'e@mail.invalid', 'test').save()
        server = self.server_proxy('test', 'test')
        try:
            server.scheduler.submit_job("{}")
        except xmlrpclib.Fault as f:
            self.assertEqual(403, f.faultCode)
        else:
            self.fail("fault not raised")

    def test_submit_job_sets_definition(self):
        user = User.objects.create_user('test', 'e@mail.invalid', 'test')
        user.user_permissions.add(
            Permission.objects.get(codename='add_testjob'))
        user.save()
        server = self.server_proxy('test', 'test')
        definition = self.factory.make_job_json()
        job_id = server.scheduler.submit_job(definition)
        job = TestJob.objects.get(id=job_id)
        self.assertEqual(definition, job.definition)

    def test_cancel_job_rejects_anonymous(self):
        job = self.factory.make_testjob()
        server = self.server_proxy()
        try:
            server.scheduler.cancel_job(job.id)
        except xmlrpclib.Fault as f:
            self.assertEqual(401, f.faultCode)
        else:
            self.fail("fault not raised")

    def test_cancel_job_rejects_unpriv_user(self):
        job = self.factory.make_testjob()
        User.objects.create_user('test', 'e@mail.invalid', 'test').save()
        server = self.server_proxy('test', 'test')
        try:
            server.scheduler.cancel_job(job.id)
        except xmlrpclib.Fault as f:
            self.assertEqual(403, f.faultCode)
        else:
            self.fail("fault not raised")

    def test_cancel_job_cancels_job(self):
        user = User.objects.create_user('test', 'e@mail.invalid', 'test')
        user.save()
        job = self.factory.make_testjob(submitter=user)
        server = self.server_proxy('test', 'test')
        server.scheduler.cancel_job(job.id)
        job = TestJob.objects.get(pk=job.pk)
        self.assertEqual(TestJob.CANCELED, job.status)


class TransactionTestCaseWithFactory(TransactionTestCase):

    def setUp(self):
        TransactionTestCase.setUp(self)
        self.factory = ModelFactory()


class NonthreadedDatabaseJobSource(DatabaseJobSource):
    deferToThread = staticmethod(lambda f, *args, **kw: f(*args, **kw))


class TestDBJobSource(TransactionTestCaseWithFactory):

    def setUp(self):
        super(TestDBJobSource, self).setUp()
        self.source = NonthreadedDatabaseJobSource()
        # The lava-health user is created by a migration in production
        # databases, but removed from the test database by the django
        # machinery.
        User.objects.create_user(
            username='lava-health', email='lava@lava.invalid')

    def test_getBoardList(self):
        self.factory.make_device(hostname='panda01')
        self.assertEqual(['panda01'], self.source.getBoardList())

    def test_getJobForBoard_returns_json(self):
        device = self.factory.make_device(hostname='panda01')
        definition = self.factory.make_job_data(target='panda01')
        self.factory.make_testjob(
            requested_device=device, definition=json.dumps(definition))
        self.assertEqual(
            definition, self.source.getJobForBoard('panda01'))

    @property
    def health_job(self):
        return self.factory.make_job_json(health_check=True)

    @property
    def ordinary_job(self):
        return self.factory.make_job_json(health_check=False)

    def assertHealthJobAssigned(self, device):
        job_data = self.source.getJobForBoard(device.hostname)
        if job_data is None:
            self.fail("no job assigned")
        self.assertTrue(
            job_data['health_check'],
            'getJobForBoard did not return health check job')

    def assertHealthJobNotAssigned(self, device):
        job_data = self.source.getJobForBoard(device.hostname)
        if job_data is None:
            self.fail("no job assigned")
        self.assertFalse(
            job_data['health_check'],
            'getJobForBoard returned health check job')

    def test_getJobForBoard_returns_health_check_if_no_last_health_job(self):
        device_type = self.factory.make_device_type(
            health_check_job=self.health_job)
        device = self.factory.make_device(
            device_type=device_type, health_status=Device.HEALTH_PASS)
        self.factory.make_testjob(
            requested_device=device, definition=self.ordinary_job)
        self.assertHealthJobAssigned(device)

    def test_getJobForBoard_returns_health_check_if_old_last_health_job(self):
        device_type = self.factory.make_device_type(
            health_check_job=self.health_job)
        device = self.factory.make_device(
            device_type=device_type, health_status=Device.HEALTH_PASS,
            last_health_report_job=self.factory.make_testjob(
                end_time=datetime.datetime.now() - datetime.timedelta(weeks=1)))
        self.factory.make_testjob(
            requested_device=device, definition=self.ordinary_job)
        self.assertHealthJobAssigned(device)

    def test_getJobForBoard_returns_job_if_healthy_and_last_health_job_recent(self):
        device_type = self.factory.make_device_type(
            health_check_job=self.health_job)
        device = self.factory.make_device(
            device_type=device_type, health_status=Device.HEALTH_PASS,
            last_health_report_job=self.factory.make_testjob(
                end_time=datetime.datetime.now() - datetime.timedelta(hours=1)))
        self.factory.make_testjob(
            requested_device=device, definition=self.ordinary_job)
        self.assertHealthJobNotAssigned(device)

    def test_getJobForBoard_returns_health_check_if_health_unknown(self):
        device_type = self.factory.make_device_type(
            health_check_job=self.health_job)
        device = self.factory.make_device(
            device_type=device_type, health_status=Device.HEALTH_UNKNOWN,
            last_health_report_job=self.factory.make_testjob(
                end_time=datetime.datetime.now() - datetime.timedelta(hours=1)))
        self.factory.make_testjob(
            requested_device=device, definition=self.ordinary_job)
        self.assertHealthJobAssigned(device)

    def test_getJobForBoard_returns_None_if_no_job(self):
        self.factory.make_device(hostname='panda01')
        self.assertEqual(
            None, self.source.getJobForBoard('panda01'))

    def test_getJobForBoard_considers_device_type(self):
        panda_type = self.factory.ensure_device_type(name='panda')
        self.factory.make_device(hostname='panda01', device_type=panda_type)
        definition = self.factory.make_job_data()
        self.factory.make_testjob(
            requested_device_type=panda_type,
            definition=json.dumps(definition))
        definition['target'] = 'panda01'
        self.assertEqual(
            definition, self.source.getJobForBoard('panda01'))

    def test_getJobForBoard_prefers_older(self):
        panda_type = self.factory.ensure_device_type(name='panda')
        panda01 = self.factory.make_device(
            hostname='panda01', device_type=panda_type)
        first_definition = self.factory.make_job_data(foo='bar', target='panda01')
        second_definition = self.factory.make_job_data(foo='baz', target='panda01')
        self.factory.make_testjob(
            requested_device=panda01, definition=json.dumps(first_definition),
            submit_time=datetime.datetime.now() - datetime.timedelta(days=1))
        self.factory.make_testjob(
            requested_device=panda01, definition=json.dumps(second_definition),
            submit_time=datetime.datetime.now())
        self.assertEqual(
            first_definition,
            self.source.getJobForBoard('panda01'))

    def test_getJobForBoard_prefers_directly_targeted(self):
        panda_type = self.factory.ensure_device_type(name='panda')
        panda01 = self.factory.make_device(
            hostname='panda01', device_type=panda_type)
        type_definition = self.factory.make_job_data()
        self.factory.make_testjob(
            requested_device_type=panda_type,
            definition=json.dumps(type_definition),
            submit_time=datetime.datetime.now() - datetime.timedelta(days=1))
        device_definition = self.factory.make_job_data(
            foo='baz', target='panda01')
        self.factory.make_testjob(
            requested_device=panda01,
            definition=json.dumps(device_definition))
        self.assertEqual(
            device_definition,
            self.source.getJobForBoard('panda01'))

    def test_getJobForBoard_avoids_targeted_to_other_board_of_same_type(self):
        panda_type = self.factory.ensure_device_type(name='panda')
        panda01 = self.factory.make_device(
            hostname='panda01', device_type=panda_type)
        self.factory.make_device(hostname='panda02', device_type=panda_type)
        definition = self.factory.make_job_data(foo='bar', target='panda01')
        self.factory.make_testjob(
            requested_device=panda01,
            definition=json.dumps(definition))
        self.assertEqual(
            None,
            self.source.getJobForBoard('panda02'))

    def _makeBoardWithTags(self, tags):
        board = self.factory.make_device()
        for tag_name in tags:
            board.tags.add(Tag.objects.get_or_create(name=tag_name)[0])
        return board

    def _makeJobWithTagsForBoard(self, tags, board):
        job = self.factory.make_testjob(requested_device=board)
        for tag_name in tags:
            job.tags.add(Tag.objects.get_or_create(name=tag_name)[0])
        return job

    def assertBoardWithTagsGetsJobWithTags(self, board_tags, job_tags):
        board = self._makeBoardWithTags(board_tags)
        self._makeJobWithTagsForBoard(job_tags, board)
        self.assertEqual(
            board.hostname,
            self.source.getJobForBoard(board.hostname)['target'])

    def assertBoardWithTagsDoesNotGetJobWithTags(self, board_tags, job_tags):
        board = self._makeBoardWithTags(board_tags)
        self._makeJobWithTagsForBoard(job_tags, board)
        self.assertEqual(
            None,
            self.source.getJobForBoard(board.hostname))

    def test_getJobForBoard_does_not_return_job_if_board_lacks_tag(self):
        self.assertBoardWithTagsDoesNotGetJobWithTags([], ['tag'])

    def test_getJobForBoard_returns_job_if_board_has_tag(self):
        self.assertBoardWithTagsGetsJobWithTags(['tag'], ['tag'])

    def test_getJobForBoard_returns_job_if_board_has_both_tags(self):
        self.assertBoardWithTagsGetsJobWithTags(['tag1', 'tag2'], ['tag1', 'tag2'])

    def test_getJobForBoard_returns_job_if_board_has_extra_tags(self):
        self.assertBoardWithTagsGetsJobWithTags(['tag1', 'tag2'], ['tag1'])

    def test_getJobForBoard_does_not_return_job_if_board_has_only_one_tag(self):
        self.assertBoardWithTagsDoesNotGetJobWithTags(['tag1'], ['tag1', 'tag2'])

    def test_getJobForBoard_does_not_return_job_if_board_has_unrelated_tag(self):
        self.assertBoardWithTagsDoesNotGetJobWithTags(['tag1'], ['tag2'])

    def test_getJobForBoard_does_not_return_job_if_only_one_tag_matches(self):
        self.assertBoardWithTagsDoesNotGetJobWithTags(['tag1', 'tag2'], ['tag1', 'tag3'])

    def test_getJobForBoard_sets_start_time(self):
        device = self.factory.make_device(hostname='panda01')
        job = self.factory.make_testjob(requested_device=device)
        before = datetime.datetime.now()
        self.source.getJobForBoard('panda01')
        after = datetime.datetime.now()
        # reload from the database
        job = TestJob.objects.get(pk=job.pk)
        self.assertTrue(before < job.start_time < after)

    def test_getJobForBoard_set_statuses(self):
        device = self.factory.make_device(hostname='panda01')
        job = self.factory.make_testjob(requested_device=device)
        self.source.getJobForBoard('panda01')
        # reload from the database
        job = TestJob.objects.get(pk=job.pk)
        device = Device.objects.get(pk=device.pk)
        self.assertEqual(
            (Device.RUNNING, TestJob.RUNNING),
            (device.status, job.status))

    def test_getJobForBoard_sets_running_job(self):
        device = self.factory.make_device(hostname='panda01')
        job = self.factory.make_testjob(requested_device=device)
        self.source.getJobForBoard('panda01')
        # reload from the database
        job = TestJob.objects.get(pk=job.pk)
        device = Device.objects.get(pk=device.pk)
        self.assertEqual(job, device.current_job)

    def test_getJobForBoard_creates_token(self):
        device = self.factory.make_device(hostname='panda01')
        job = self.factory.make_testjob(requested_device=device)
        self.source.getJobForBoard('panda01')
        # reload from the database
        job = TestJob.objects.get(pk=job.pk)
        device = Device.objects.get(pk=device.pk)
        self.assertIsNotNone(job.submit_token)
        self.assertEqual(job.submitter, job.submit_token.user)

    def test_getJobForBoard_inserts_target_into_json(self):
        panda_type = self.factory.ensure_device_type(name='panda')
        self.factory.make_device(hostname='panda01', device_type=panda_type)
        definition = self.factory.make_job_data(device_type='panda')
        self.factory.make_testjob(
            requested_device_type=panda_type,
            definition=json.dumps(definition))
        json_data = self.source.getJobForBoard('panda01')
        self.assertIn('target', json_data)
        self.assertEqual('panda01', json_data['target'])

    def test_getJobForBoard_inserts_submit_token_into_json(self):
        panda_type = self.factory.ensure_device_type(name='panda')
        self.factory.make_device(hostname='panda01', device_type=panda_type)
        definition = {
            'actions': [
                {
                    "command": "submit_results",
                    "parameters":
                    {
                        "server": "http://test-server/RPC2/",
                        "stream": "/private/personal/test/test/",
                    }
                }
            ]
        }
        job = self.factory.make_testjob(
            requested_device_type=panda_type,
            definition=json.dumps(definition))
        json_data = self.source.getJobForBoard('panda01')
        job = TestJob.objects.get(pk=job.pk)
        submit_job_params = json_data['actions'][0]['parameters']
        self.assertIn('token', submit_job_params)
        self.assertEqual(job.submit_token.secret, submit_job_params['token'])

    def test_getJobForBoard_adds_user_to_url(self):
        panda_type = self.factory.ensure_device_type(name='panda')
        self.factory.make_device(hostname='panda01', device_type=panda_type)
        user = User.objects.create_user('test', 'e@mail.invalid', 'test')
        user.save()
        definition = {
            'actions': [
                {
                    "command": "submit_results",
                    "parameters":
                    {
                        "server": "http://test-server/RPC2/",
                        "stream": "/private/personal/test/test/",
                    }
                }
            ]
        }
        job = self.factory.make_testjob(
            requested_device_type=panda_type, submitter=user,
            definition=json.dumps(definition))
        json_data = self.source.getJobForBoard('panda01')
        job = TestJob.objects.get(pk=job.pk)
        submit_job_params = json_data['actions'][0]['parameters']
        self.assertEqual("http://test@test-server/RPC2/", submit_job_params['server'])

    def get_device_and_running_job(self):
        device = self.factory.make_device(hostname='panda01')
        job = self.factory.make_testjob(requested_device=device)
        self.source.getJobForBoard('panda01')
        return device, TestJob.objects.get(pk=job.pk)

    def test_jobCompleted_set_statuses_success(self):
        device, job = self.get_device_and_running_job()
        self.source.jobCompleted('panda01', 0)
        job = TestJob.objects.get(pk=job.pk)
        device = Device.objects.get(pk=device.pk)
        self.assertEqual(
            (Device.IDLE, TestJob.COMPLETE),
            (device.status, job.status))

    def test_jobCompleted_set_statuses_failure(self):
        device, job = self.get_device_and_running_job()
        self.source.jobCompleted('panda01', 1)
        job = TestJob.objects.get(pk=job.pk)
        device = Device.objects.get(pk=device.pk)
        self.assertEqual(
            (Device.IDLE, TestJob.INCOMPLETE),
            (device.status, job.status))

    def test_jobCompleted_works_on_device_type_targeted(self):
        device = self.factory.make_device(hostname='panda01')
        job = self.factory.make_testjob(
            requested_device_type=device.device_type)
        self.source.getJobForBoard('panda01')
        self.source.jobCompleted('panda01', 0)
        job = TestJob.objects.get(pk=job.pk)
        device = Device.objects.get(pk=device.pk)
        self.assertEqual(
            (Device.IDLE, TestJob.COMPLETE),
            (device.status, job.status))

    def test_jobCompleted_sets_end_time(self):
        device, job = self.get_device_and_running_job()
        before = datetime.datetime.now()
        self.source.jobCompleted('panda01', 0)
        after = datetime.datetime.now()
        job = TestJob.objects.get(pk=job.pk)
        self.assertTrue(before < job.end_time < after)

    def test_jobCompleted_clears_current_job(self):
        device, job = self.get_device_and_running_job()
        self.source.jobCompleted('panda01', 0)
        device = Device.objects.get(pk=device.pk)
        self.assertEquals(None, device.current_job)

    def test_jobCompleted_deletes_token(self):
        device, job = self.get_device_and_running_job()
        token = job.submit_token
        self.source.jobCompleted('panda01', 0)
        self.assertRaises(
            AuthToken.DoesNotExist,
            AuthToken.objects.get, pk=token.pk)

    def test_getLogFileForJobOnBoard_returns_writable_file(self):
        device, job = self.get_device_and_running_job()
        definition = {'foo': 'bar'}
        self.factory.make_testjob(
            requested_device=device, definition=json.dumps(definition))
        log_file = self.source.getLogFileForJobOnBoard('panda01')
        log_file.write('a')
        log_file.close()



import logging

from twisted.application.service import Service
from twisted.internet import defer
from twisted.internet.task import LoopingCall

from lava_scheduler_daemon.board import Board, catchall_errback


class BoardSet(Service):

    def __init__(self, source, dispatcher, reactor, daemon_options):
        self.logger = logging.getLogger(__name__ + '.BoardSet')
        self.source = source
        self.boards = {}
        self.dispatcher = dispatcher
        self.reactor = reactor
        self.daemon_options = daemon_options
        self._update_boards_call = LoopingCall(self._updateBoards)
        self._update_boards_call.clock = reactor

    def _updateBoards(self):
        self.logger.debug("Refreshing board list")
        return self.source.getBoardList().addCallback(
            self._cbUpdateBoards).addErrback(catchall_errback(self.logger))

    def _cbUpdateBoards(self, board_names):
        if set(board_names) == set(self.boards):
            return
        self.logger.info("New board list %s", board_names)
        new_boards = {}
        for board_name in board_names:
            if board_name in self.boards:
                new_boards[board_name] = self.boards.pop(board_name)
            else:
                new_boards[board_name] = Board(
                    self.source, board_name, self.dispatcher, self.reactor,
                    self.daemon_options)
                new_boards[board_name].start()
        for board in self.boards.values():
            board.stop()
        self.boards = new_boards

    def startService(self):
        self._update_boards_call.start(20)

    def stopService(self):
        self._update_boards_call.stop()
        ds = []
        dead_boards = []
        for board in self.boards.itervalues():
            ds.append(board.stop().addCallback(dead_boards.append))
        self.logger.info(
            "waiting for %s boards", len(self.boards) - len(dead_boards))
        return defer.gatherResults(ds)


